/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"
#include "UserConnection.h"
#include "ClientManager.h"
#include "ResourceManager.h"

#include "StringTokenizer.h"
#include "ConnectionManager.h"

const string UserConnection::FEATURE_MINISLOTS = "MiniSlots";
const string UserConnection::FEATURE_XML_BZLIST = "XmlBZList";
const string UserConnection::FEATURE_ADCGET = "ADCGet";
const string UserConnection::FEATURE_ZLIB_GET = "ZLIG";
const string UserConnection::FEATURE_TTHL = "TTHL";
const string UserConnection::FEATURE_TTHF = "TTHF";

const string UserConnection::FILE_NOT_AVAILABLE = "File Not Available";

const string Transfer::TYPE_FILE = "file";
const string Transfer::TYPE_TTHL = "tthl";

const string Transfer::USER_LIST_NAME_BZ = "files.xml.bz2";

const string UserConnection::UPLOAD = "Upload";
const string UserConnection::DOWNLOAD = "Download";

Transfer::Transfer(UserConnection& conn) : start(0), lastTick(GET_TICK()), runningAverage(0),
last(0), actual(0), pos(0), startPos(0), size(-1), userConnection(conn), tth(NULL) { }

void Transfer::updateRunningAverage() {
	uint64_t tick = GET_TICK();
	if(tick > (lastTick + 250)) {
		uint64_t diff = tick - lastTick;
		int64_t tot = getTotal();
		if(((tick - getStart()) < AVG_PERIOD)) {
			runningAverage = getAverageSpeed();
		} else {
			int64_t bdiff = tot - last;
			int64_t avg = bdiff * (int64_t)1000 / diff;
			if(diff > AVG_PERIOD) {
				runningAverage = avg;
			} else {
				// Weighted average...
				runningAverage = ((avg * diff) + (runningAverage*(AVG_PERIOD-diff)))/AVG_PERIOD;
			}
		}
		last = tot;
	}
	lastTick = tick;
}

void Transfer::getParams(const UserConnection& aSource, StringMap& params) {
	params["userNI"] = Util::toString(ClientManager::getInstance()->getNicks(aSource.getUser()->getCID()));
	params["userI4"] = aSource.getRemoteIp();
	StringList hubNames = ClientManager::getInstance()->getHubNames(aSource.getUser()->getCID());
	if(hubNames.empty())
		hubNames.push_back(STRING(OFFLINE));
	params["hub"] = Util::toString(hubNames);
	StringList hubs = ClientManager::getInstance()->getHubs(aSource.getUser()->getCID());
	if(hubs.empty())
		hubs.push_back(STRING(OFFLINE));
	params["hubURL"] = Util::toString(hubs);
	params["fileSI"] = Util::toString(getSize());
	params["fileSIshort"] = Util::formatBytes(getSize());
	params["fileSIchunk"] = Util::toString(getTotal());
	params["fileSIchunkshort"] = Util::formatBytes(getTotal());
	params["fileSIactual"] = Util::toString(getActual());
	params["fileSIactualshort"] = Util::formatBytes(getActual());
	params["speed"] = Util::formatBytes(getAverageSpeed()) + "/s";
	params["time"] = Util::formatSeconds((GET_TICK() - getStart()) / 1000);
	if(getTTH() != NULL)
		params["fileTR"] = getTTH()->toBase32();
}

User::Ptr Transfer::getUser() {
	return getUserConnection().getUser();
}

void UserConnection::onLine(char* aLine, int iLineLen) throw() {
	if(aLine[0] == '$') {
		switch(aLine[1]) {
	        case 'M':
                if(iLineLen < 9) {
                    break;
                }

	        	// $MyNick
	        	if(memcmp(aLine+2, "yNick ", 6) == NULL) {
	                string sNick = string(aLine+8, iLineLen-8);
	        		fire(UserConnectionListener::MyNick(), this, Text::toUtf8(sNick));
	                return;
                }

                // $MaxedOut
	        	if(memcmp(aLine+2, "axedOut", 7) == NULL) {
	        		fire(UserConnectionListener::MaxedOut(), this);
	        		return;
	            }
	            
	            break;
	    	case 'D':
	            // $Direction
	            if(iLineLen > 11 && memcmp(aLine+2, "irection ", 9) == NULL) {
                    char *sTemp = strchr(aLine+11, ' ');
                    if(sTemp == NULL || (sTemp-aLine)-11 == 0 || iLineLen-(sTemp-aLine)-1 == 0)
                        return;
					
					string sDir = string(aLine+11, (sTemp-aLine)-11);
					string sNum = string(sTemp+1, iLineLen-(sTemp-aLine)-1);
                    fire(UserConnectionListener::Direction(), this, sDir, sNum);

	        		return;
	            }
	            
	            break;
	    	case 'E':
	    		// $Error
	            if(iLineLen > 7 && memcmp(aLine+2, "rror ", 5) == NULL) {
	        		if(_stricmp(aLine+7, FILE_NOT_AVAILABLE.c_str()) == 0 || 
	        			strstr(aLine+7, /*path/file*/" no more exists") != 0) {
	        			fire(UserConnectionListener::FileNotAvailable(), this);
	        		} else {
                        string sErr = string(aLine+7, iLineLen-7);
						fire(UserConnectionListener::Failed(), this, Text::toUtf8(sErr));
	        		}
	        		return;
	            }
	            
	            break;
	    	case 'F':
	    		// $FileLength
	            if(iLineLen > 12 && memcmp(aLine+2, "ileLength ", 10) == NULL) {
	        		fire(UserConnectionListener::FileLength(), this, _atoi64(aLine+12));
	        		return;
	            }
	            
	            break;
	    	case 'G':
	            // $Get
	        	if(iLineLen > 5 && memcmp(aLine+2, "et ", 3) == NULL) {
					char *sTemp = strrchr(aLine+5, '$');
                    if(sTemp == NULL || (sTemp-aLine)-5 == 0 || iLineLen-(sTemp-aLine)-1 == 0)
                        return;

                    string sFile = string(aLine+5, (sTemp-aLine)-5);
	        		fire(UserConnectionListener::Get(), this, Text::toUtf8(sFile), _atoi64(sTemp+1) - (int64_t)1);
	        		return;
	            }
	            
	            // $GetListLen
	            if(iLineLen == 11 && memcmp(aLine+2, "etListLen", 9) == NULL) {
	        		fire(UserConnectionListener::GetListLength(), this);
	        		return;
	            }
	            
	            break;
	    	case 'U':            
	            // $UGetBlock
	        	if(iLineLen > 11 && memcmp(aLine+2, "GetBlock ", 9) == NULL) {
                    char *sBlockParts[] = { NULL, NULL, NULL };
                    unsigned short iBlockPartsLen[] = { 0, 0, 0 };
                    
                    unsigned char cPart = 0;
                    
                    sBlockParts[cPart] = aLine+11;
                    
                    for(int i = 11; i < iLineLen; i++) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                            iBlockPartsLen[cPart] = (unsigned short)((aLine+i)-sBlockParts[cPart]);
                            
                            if(cPart == 1) {
                                sBlockParts[2] = aLine+i+1;
                                iBlockPartsLen[2] = (unsigned short)(iLineLen-i-1);
								break;
                            }
                            
                            cPart++;
                            sBlockParts[cPart] = aLine+i+1;
                        }
                    }

					if(iBlockPartsLen[0] == 0 || iBlockPartsLen[1] == 0 || iBlockPartsLen[2] == 0)
						return;

                	int64_t start = _atoi64(sBlockParts[0]);
                	if(start < 0) {
                		disconnect();
                		return;
                	}
                
                	int64_t bytes = _atoi64(sBlockParts[1]);
                	string sFile = string(sBlockParts[2], iBlockPartsLen[2]);
                	fire(UserConnectionListener::GetBlock(), this, Text::toUtf8(sFile), start, bytes);
	        		
	        		return;
	            }
	            
	            break;
	    	case 'K':
	            // $Key
	            if(iLineLen > 5 && memcmp(aLine+2, "ey ", 3) == NULL) {
	                string sKey = string(aLine+5, iLineLen-5);
	        		fire(UserConnectionListener::Key(), this, sKey);
	        		return;
	            }
	            
	            break;
	    	case 'L':
	            // $Lock
	            if(iLineLen > 6 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"ock ")[0]) {
                    char *sTemp = (char *)memchr(aLine+6, ' ', iLineLen-6);				
	        		if(sTemp == NULL || (sTemp-aLine)-6 == 0 || iLineLen-(sTemp-aLine)-1 == 0) {
                        string sLock = string(aLine+6, iLineLen-6);
						fire(UserConnectionListener::CLock(), this, sLock, Util::emptyString);
	        		} else {
                        string sLock = string(aLine+6, (sTemp-aLine)-6);
                        string sPk = string(sTemp+1, iLineLen-(sTemp-aLine)-1);
						fire(UserConnectionListener::CLock(), this, sLock, sPk);
	        		}
	        		return;
	            }
	            
	            break;
	    	case 'S':
	            // $Send
	            if(iLineLen == 5 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"end\0")[0]) {
	        		fire(UserConnectionListener::Send(), this);
	        		return;
	            }
	            
	            // $Sending
	        	if(iLineLen > 9 && memcmp(aLine+2, "ending ", 7) == NULL) {
	        		fire(UserConnectionListener::Sending(), this, _atoi64(aLine+9));
	        		return;
	            }
	            
	            // $Supports
	        	if(iLineLen > 10 && ((uint64_t *)(aLine+2))[0] == ((uint64_t *)"upports ")[0]) {
					vector<string> lst;

                    char *sSupport = aLine+10;
                    
                    for(int i = 10; i < iLineLen; i++) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                        } else if(i != iLineLen-1) {
                            continue;
                        } else {
                            i++;
                        }
                        
                        string sSupp = string(sSupport, (aLine+i)-sSupport);
						lst.push_back(sSupp);
                        
                        sSupport = aLine+i+1;
                    }
	
	        		fire(UserConnectionListener::Supports(), this, lst);
	        		return;
	            }
	            
	            break;
	    	case 'A':
                // $ADCGET / $ADCSND
	            if(iLineLen > 7 && ((uint16_t *)(aLine+2))[0] == ((uint16_t *)"DC")[0]) {
                    string sAnihilateDirectConnectGet = string(aLine, iLineLen);
	        		dispatch(sAnihilateDirectConnectGet, true);
	        		return;
	            }
	            
	            break;
			default:
				break;
		}
	}

	// We shouldn't be here?
	dcdebug("Unknown DC command: %.50s\n", aLine);
}

void UserConnection::connect(const string& aServer, uint16_t aPort) throw(SocketException, ThreadException) { 
	dcassert(!socket);
		
	socket = BufferedSocket::getSocket('|');
	socket->addListener(this);
	socket->connect(aServer, aPort, true);
}

void UserConnection::accept(const Socket& aServer) throw(SocketException, ThreadException) {
	dcassert(!socket);
	socket = BufferedSocket::getSocket('|');
	socket->addListener(this);
	socket->accept(aServer);
}

void UserConnection::on(BufferedSocketListener::Failed, const string& aLine) throw() {
	setState(STATE_UNCONNECTED);
	fire(UserConnectionListener::Failed(), this, aLine);
	
	delete this;
}

/* Some parts Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org */
